from dataclasses import dataclass
from typing import TYPE_CHECKING, Collection, List, Literal, Optional, Type, Union

from lonelypsp.sync_io import SyncReadableBytesIO
from lonelypsp.stateful.constants import (
    PubSubStatefulMessageFlags,
    SubscriberToBroadcasterStatefulMessageType,
)
from lonelypsp.compat import fast_dataclass
from lonelypsp.stateful.generic_parser import S2B_MessageParser
from lonelypsp.stateful.parser_helpers import parse_simple_headers
from lonelypsp.stateful.serializer_helpers import (
    MessageSerializer,
    serialize_simple_message,
)


@fast_dataclass
class S2B_UnsubscribeExact:
    """
    S2B = Subscriber to Broadcaster
    See the type enum documentation for more information on the fields
    """

    type: Literal[SubscriberToBroadcasterStatefulMessageType.UNSUBSCRIBE_EXACT]
    """discriminator value"""

    authorization: Optional[str]
    """url: websocket:<nonce>:<ctr>
    
    an empty string is reinterpreted as None for consistency between
    minimal headers mode and expanded headers mode
    """

    topic: bytes
    """the topic to unsubscribe from"""


_exact_headers: Collection[str] = ("authorization", "x-topic")


class S2B_UnsubscribeExactParser:
    """Satisfies S2B_MessageParser[S2B_UnsubscribeExact]"""

    @classmethod
    def relevant_types(cls) -> List[SubscriberToBroadcasterStatefulMessageType]:
        return [SubscriberToBroadcasterStatefulMessageType.UNSUBSCRIBE_EXACT]

    @classmethod
    def parse(
        cls,
        flags: PubSubStatefulMessageFlags,
        type: SubscriberToBroadcasterStatefulMessageType,
        payload: SyncReadableBytesIO,
    ) -> S2B_UnsubscribeExact:
        assert type == SubscriberToBroadcasterStatefulMessageType.UNSUBSCRIBE_EXACT

        headers = parse_simple_headers(flags, payload, _exact_headers)
        authorization_bytes = headers.get("authorization", b"")
        authorization = (
            None if not authorization_bytes else authorization_bytes.decode("utf-8")
        )

        topic = headers["x-topic"]
        return S2B_UnsubscribeExact(
            type=type,
            authorization=authorization,
            topic=topic,
        )


if TYPE_CHECKING:
    _: Type[S2B_MessageParser[S2B_UnsubscribeExact]] = S2B_UnsubscribeExactParser


def serialize_s2b_unsubscribe_exact(
    msg: S2B_UnsubscribeExact, /, *, minimal_headers: bool
) -> Union[bytes, bytearray, memoryview]:
    """Satisfies MessageSerializer[S2B_UnsubscribeExact]"""
    return serialize_simple_message(
        type=msg.type,
        header_names=_exact_headers,
        header_values=(
            b"" if msg.authorization is None else msg.authorization.encode("utf-8"),
            msg.topic,
        ),
        minimal_headers=minimal_headers,
        payload=b"",
    )


if TYPE_CHECKING:
    __: MessageSerializer[S2B_UnsubscribeExact] = serialize_s2b_unsubscribe_exact


@fast_dataclass
class S2B_UnsubscribeGlob:
    """
    S2B = Subscriber to Broadcaster
    See the type enum documentation for more information on the fields
    """

    type: Literal[SubscriberToBroadcasterStatefulMessageType.UNSUBSCRIBE_GLOB]
    """discriminator value"""

    authorization: Optional[str]
    """url: websocket:<nonce>:<ctr>
    
    an empty string is reinterpreted as None for consistency between
    minimal headers mode and expanded headers mode
    """

    glob: str
    """the glob pattern to unsubscribe from"""


_glob_headers: Collection[str] = ("authorization", "x-glob")


class S2B_UnsubscribeGlobParser:
    """Satisfies S2B_MessageParser[S2B_UnsubscribeGlob]"""

    @classmethod
    def relevant_types(cls) -> List[SubscriberToBroadcasterStatefulMessageType]:
        return [SubscriberToBroadcasterStatefulMessageType.UNSUBSCRIBE_GLOB]

    @classmethod
    def parse(
        cls,
        flags: PubSubStatefulMessageFlags,
        type: SubscriberToBroadcasterStatefulMessageType,
        payload: SyncReadableBytesIO,
    ) -> S2B_UnsubscribeGlob:
        assert type == SubscriberToBroadcasterStatefulMessageType.UNSUBSCRIBE_GLOB

        headers = parse_simple_headers(flags, payload, _glob_headers)
        authorization_bytes = headers.get("authorization", b"")
        authorization = (
            None if not authorization_bytes else authorization_bytes.decode("utf-8")
        )

        glob = headers["x-glob"].decode("utf-8")
        return S2B_UnsubscribeGlob(
            type=type,
            authorization=authorization,
            glob=glob,
        )


if TYPE_CHECKING:
    ___: Type[S2B_MessageParser[S2B_UnsubscribeGlob]] = S2B_UnsubscribeGlobParser


def serialize_s2b_unsubscribe_glob(
    msg: S2B_UnsubscribeGlob, /, *, minimal_headers: bool
) -> Union[bytes, bytearray, memoryview]:
    """Satisfies MessageSerializer[S2B_UnsubscribeGlob]"""
    return serialize_simple_message(
        type=msg.type,
        header_names=_glob_headers,
        header_values=(
            b"" if msg.authorization is None else msg.authorization.encode("utf-8"),
            msg.glob.encode("utf-8"),
        ),
        minimal_headers=minimal_headers,
        payload=b"",
    )


if TYPE_CHECKING:
    ____: MessageSerializer[S2B_UnsubscribeGlob] = serialize_s2b_unsubscribe_glob


S2B_Unsubscribe = Union[S2B_UnsubscribeExact, S2B_UnsubscribeGlob]



from typing import TYPE_CHECKING, Collection, List, Literal, Type, Union

from lonelypsp.sync_io import SyncReadableBytesIO
from lonelypsp.stateful.constants import (
    BroadcasterToSubscriberStatefulMessageType,
    PubSubStatefulMessageFlags,
)
from lonelypsp.compat import fast_dataclass
from lonelypsp.stateful.generic_parser import B2S_MessageParser
from lonelypsp.stateful.parser_helpers import parse_simple_headers
from lonelypsp.stateful.serializer_helpers import (
    MessageSerializer,
    int_to_minimal_unsigned,
    serialize_simple_message,
)


@fast_dataclass
class B2S_EnableZstdCustom:
    """
    B2S = Broadcaster to Subscriber
    See the type enum documentation for more information on the fields
    """

    type: Literal[BroadcasterToSubscriberStatefulMessageType.ENABLE_ZSTD_CUSTOM]
    """discriminator value"""

    identifier: int
    """the identifier the broadcaster has assigned to compressing with this
    dictionary
    """

    compression_level: int
    """the compression level (any negative integer up to and including positive 22)
    that the broadcaster recommends for this dictionary; the subscriber is free to
    ignore this recommendation
    """

    min_size: int
    """the minimum in size in bytes that the broadcaster recommends for using
    this preset; the subscriber is free to ignore this recommendation
    """

    max_size: int
    """the maximum in size in bytes that the broadcaster recommends for using
    this preset; the subscriber is free to ignore this recommendation. 2**64-1
    for no limit
    """

    dictionary: bytes
    """the compression dictionary, in bytes, that is referenced when compressing
    with this identifier
    """


_headers: Collection[str] = (
    "x-identifier",
    "x-compression-level",
    "x-min-size",
    "x-max-size",
)


class B2S_EnableZstdCustomParser:
    """Satisfies B2S_MessageParser[B2S_EnableZstdCustom]"""

    @classmethod
    def relevant_types(cls) -> List[BroadcasterToSubscriberStatefulMessageType]:
        return [BroadcasterToSubscriberStatefulMessageType.ENABLE_ZSTD_CUSTOM]

    @classmethod
    def parse(
        cls,
        flags: PubSubStatefulMessageFlags,
        type: BroadcasterToSubscriberStatefulMessageType,
        payload: SyncReadableBytesIO,
    ) -> B2S_EnableZstdCustom:
        assert type == BroadcasterToSubscriberStatefulMessageType.ENABLE_ZSTD_CUSTOM

        headers = parse_simple_headers(flags, payload, _headers)
        identifier_bytes = headers["x-identifier"]
        if len(identifier_bytes) > 8:
            raise ValueError("x-identifier must be at most 8 bytes")

        identifier = int.from_bytes(identifier_bytes, "big")

        compression_level_bytes = headers["x-compression-level"]
        if len(compression_level_bytes) > 2:
            raise ValueError("x-compression-level max 2 bytes")

        compression_level = int.from_bytes(compression_level_bytes, "big", signed=True)

        min_size_bytes = headers["x-min-size"]
        if len(min_size_bytes) > 4:
            raise ValueError("x-min-size max 4 bytes")

        min_size = int.from_bytes(min_size_bytes, "big")

        max_size_bytes = headers["x-max-size"]
        if len(max_size_bytes) > 8:
            raise ValueError("x-max-size max 8 bytes")

        max_size = int.from_bytes(max_size_bytes, "big")

        dictionary = payload.read(-1)

        return B2S_EnableZstdCustom(
            type=type,
            identifier=identifier,
            compression_level=compression_level,
            min_size=min_size,
            max_size=max_size,
            dictionary=dictionary,
        )


if TYPE_CHECKING:
    _: Type[B2S_MessageParser[B2S_EnableZstdCustom]] = B2S_EnableZstdCustomParser


def serialize_b2s_enable_zstd_custom(
    msg: B2S_EnableZstdCustom, /, *, minimal_headers: bool
) -> Union[bytes, bytearray, memoryview]:
    """Satisfies MessageSerializer[B2S_EnableZstdCustom]"""
    return serialize_simple_message(
        type=msg.type,
        header_names=_headers,
        header_values=(
            int_to_minimal_unsigned(msg.identifier),
            msg.compression_level.to_bytes(2, "big", signed=True),
            int_to_minimal_unsigned(msg.min_size),
            int_to_minimal_unsigned(msg.max_size),
        ),
        payload=msg.dictionary,
        minimal_headers=minimal_headers,
    )


if TYPE_CHECKING:
    __: MessageSerializer[B2S_EnableZstdCustom] = serialize_b2s_enable_zstd_custom



from enum import IntEnum, auto


class SubscriberToBroadcasterStatelessMessageType(IntEnum):
    """Assigns a unique integer to each type of message that a subscriber can
    send to a broadcaster over a stateless connection. The prototypical example
    of a stateless connection would be using a distinct HTTP call for each
    message.

    Because stateless connections generally already have more parsing helpers available,
    we just provide documentation on how the message should be structured, but don't
    actually provide parsers or serializers.
    """

    NOTIFY = auto()
    """The subscriber is posting a message to a specific topic
    
    ### headers
    - authorization: proof the subscriber is authorized to post to the topic

    ### body
    - 2 bytes (N): length of the topic, big-endian, unsigned
    - N bytes: the topic. if utf-8 decodable then we will attempt to match glob
      patterns, otherwise, only goes to exact subscriptions
    - 64 bytes: sha-512 hash of the message, will be rechecked
    - 8 bytes (M): length of the message, big-endian, unsigned
    - M bytes: the message
    """

    SUBSCRIBE_EXACT = auto()
    """The subscriber wants to receive messages posted to a specific topic

    ### headers
    - authorization: proof the subscriber is authorized to subscribe to the topic

    ### body
    - 2 bytes (N): length of the url that the subscriber can be reached at, big-endian, unsigned
    - N bytes: the url that the subscriber can be reached at, utf-8 encoded
    - 2 bytes (M): the length of the topic, big-endian, unsigned
    - M bytes: the topic, utf-8 encoded
    """

    SUBSCRIBE_GLOB = auto()
    """The subscriber wants to receive messages to utf-8 decodable topics which match
    a given glob pattern
    
    ### headers
    - authorization: proof the subscriber is authorized to subscribe to the pattern

    ### body
    - 2 bytes (N): length of the url that the subscriber can be reached at, big-endian, unsigned
    - N bytes: the url that the subscriber can be reached at, utf-8 encoded
    - 2 bytes (M): the length of the glob pattern, big-endian, unsigned
    - M bytes: the glob pattern, utf-8 encoded
    """

    UNSUBSCRIBE_EXACT = auto()
    """The subscriber wants to stop receiving messages posted to a specific topic

    ### headers
    - authorization: proof the subscriber is authorized to unsubscribe from the topic;
      formed exactly like the authorization header in SUBSCRIBE_EXACT
    
    ### body
    - 2 bytes (N): length of the url that the subscriber can be reached at, big-endian, unsigned
    - N bytes: the url that the subscriber can be reached at, utf-8 encoded
    - 2 bytes (M): the length of the topic, big-endian, unsigned
    - M bytes: the topic, utf-8 encoded
    """

    UNSUBSCRIBE_GLOB = auto()
    """The subscriber wants to stop receiving messages to utf-8 decodable topics which match
    a given glob pattern

    ### headers
    - authorization: proof the subscriber is authorized to unsubscribe from the pattern;
      formed exactly like the authorization header in SUBSCRIBE_GLOB

    ### body
    - 2 bytes (N): length of the url that the subscriber can be reached at, big-endian, unsigned
    - N bytes: the url that the subscriber can be reached at, utf-8 encoded
    - 2 bytes (M): the length of the glob pattern, big-endian, unsigned
    - M bytes: the glob pattern, utf-8 encoded
    """


class SubscriberToBroadcasterStatelessResponseType(IntEnum):
    """When the broadcaster reaches out to a subscriber they have the opportunity
    to respond with one of these types of messages, without headers
    """

    UNKNOWN = auto()
    """Used when no specific meaning was understood about the response"""

    UNSUBSCRIBE_IMMEDIATE = auto()
    """When sent in response to a RECEIVE message, removes whatever subscription
    caused the subscriber to receive the message

    The body is a json object with at least the following keys:
    - `unsubscribe`: the value `true`

    All other values are ignored, however, a common one worth mentioning is:
    - `reason`: a human-readable string indicating if the subscriber didn't
      like the format of the message vs just wasn't expecting a message on that
      topic vs was expecting the message but no longer wants more
    """


class BroadcasterToSubscriberStatelessMessageType(IntEnum):
    """Assigns a unique integer to each type of message that a broadcaster can
    send to a subscriber over a stateless connection. The prototypical example
    of a stateless connection would be using a distinct HTTP call for each
    message.
    """

    RECEIVE = auto()
    """The broadcaster is notifying the subscriber of a message posted to a topic
    
    ### headers
    - authorization: proof the broadcaster can notify the subscriber
    - repr-digest: contains <digest-algorithm>=<digest>[,<digest-algorithm>=<digest>...]
      where at least one of the digest algorithms is `sha512` and the digest is the
      the base64 encoded sha-512 hash of the message
    - x-topic: the topic the message was posted to

    ### body
    the message that was posted to the topic
    """
